/*
 * Cheops Next Generation GUI
 * 
 * ip_utils.c
 * Utilities for handeling IP addresses.
 *
 * Copyright(C) 1999 Brent Priddy <toopriddy@mailcity.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdarg.h>
#include <net/if.h>
#include <gtk/gtk.h>
#include <stdio.h>

#include "cheops.h"
#include "io.h"
#include "logger.h"
#include "event.h"
#include "cheops-sh.h"
#include "ip_utils.h"


#include <linux/sockios.h>  /* GLIBC don't have sockios.h? */
typedef struct {
	char ifname[17];
	struct in_addr addr;
} interfacerec;
            
typedef struct {
	struct in_addr addr;
	unsigned long naddr;    /* netmask */
	interfacerec *iface;
} routerec;

static short numinterfaces = 0, numroutes = 0;
static interfacerec *interfaces = NULL;
static routerec *routes = NULL;
                            
                            

int get_host(char *hostname, u32 *ip) 
{
	struct hostent *hp;
	hp = gethostbyname(hostname);
	if (hp) 
		memcpy(ip, hp->h_addr, sizeof(u32));
	else
		return -1;
	return 0;
}

char *ip2str(u32 ip)
{
	return inet_ntoa(*(struct in_addr *)&ip);
}

int str2ip(char *s)
{
	struct in_addr ip;
	
	inet_aton(s,&ip);
	
	return( ip.s_addr);
}

int allones(u32 mask)
{
	/* Check to be sure that this only has only a string of ones,
	   return non-zero on success */
	while(mask & 0x80000000) {
		mask = mask << 1;
	}
	return !mask;
}

unsigned short inet_checksum (void *addr, int len)
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;
        
	while (nleft > 1)
	{
		sum += *w++;
		nleft -= 2;
	}
	/* mop up an odd byte, if necessary */
	if (nleft == 1)
	{
		*(u_char *) (&answer) = *(u_char *) w;
		sum += answer;
	}
                                     	                            
	sum = (sum >> 16) + (sum & 0xffff);   /* add hi 16 to low 16 */
	sum += (sum >> 16);		/* add carry */
	answer = ~sum;		/* truncate to 16 bits */
	return (answer);
}

void sendicmp (int soc, int ttl, struct in_addr to)
{
	unsigned char pkt[8192];
	struct ip *ip = (struct ip *)pkt;
	struct icmp *icmp = (struct icmp *)&pkt[sizeof(struct ip)];
	int len;
	struct sockaddr_in dest, src;
  
	src.sin_addr = getlocalip(to.s_addr);

	src.sin_family = AF_INET;
	src.sin_port = 0;
	dest.sin_addr = to;
	dest.sin_family = AF_INET;

/*-- IP HDR --*/
	ip->ip_hl = 5;
	ip->ip_v = 4;
	ip->ip_tos = 0;
	ip->ip_id = htons(getpid());
	ip->ip_off = 0;
	ip->ip_ttl = ttl;
	ip->ip_src = src.sin_addr;
	ip->ip_dst = to;
	ip->ip_sum = 0;

    len = ICMP_MINLEN;
    dest.sin_port = 0;
    ip->ip_p = IPPROTO_ICMP;

/*-- ICMP HDR --*/
  	icmp->icmp_type = ICMP_ECHO;
  	icmp->icmp_code = 0;
  	icmp->icmp_id = htons(getpid());
  	icmp->icmp_seq = htons(ttl);

  	icmp->icmp_cksum = 0;
 	icmp->icmp_cksum = inet_checksum (icmp, len);

	ip->ip_len = htons(len + sizeof(struct ip));
	ip->ip_sum = inet_checksum ((void *) ip, sizeof (struct ip)+len);

	if (ip->ip_sum == 0)
		ip->ip_sum=0xffff;

	sendto (soc, (void *) pkt, sizeof(struct ip) + len, 0, (struct sockaddr *)&dest, sizeof (dest));
}

struct in_addr getlocalip (unsigned int dest)
{
	static struct in_addr ina;
	int i;
	
	init_route_tables();
	
	for (i = 0; i < numroutes; i++)
	{
		if ((dest & routes[i].naddr) == (unsigned long) routes[i].addr.s_addr)
		{
			return (routes[i].iface->addr);
		}
	}
	ina.s_addr = 0;
	return ina;
}

void init_route_tables(void)
{
	int ifsock, i, i1, found;
	struct ifconf ifc;
	struct ifreq *ifr;
	char buf[1024], iface[16], *ptr;
	FILE *f;

	/* Create a channel to the NET kernel. */
	if ((ifsock = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		perror ("socket");
		exit (EXIT_FAILURE);
	}

	ifc.ifc_len = sizeof (buf);
	ifc.ifc_buf = buf;
	if (ioctl (ifsock, SIOCGIFCONF, &ifc) < 0)
	{
		perror ("opening interface socket");
		close (ifsock);
		exit (EXIT_FAILURE);
	}

	numinterfaces = (ifc.ifc_len / sizeof (struct ifreq));
	interfaces = (interfacerec *) malloc (numinterfaces * sizeof (interfacerec));

	ifr = ifc.ifc_req;
	for (i = 0; i < numinterfaces; i++, ifr++)
	{
		strcpy (interfaces[i].ifname, ifr->ifr_name);
		memcpy (&interfaces[i].addr, &((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr, sizeof (struct in_addr));
		if (ioctl (ifsock, SIOCGIFADDR, ifr) < 0)
			printf ("Couldn't get address for %s\n", ifr->ifr_name);
	}
	close (ifsock);

	if ((f = fopen ("/proc/net/route", "r")) == NULL)
	{
		perror ("opening /proc/net/route");
		exit (EXIT_FAILURE);
	}

	numroutes = 0;
	fgets (buf, sizeof (buf), f);         /* strip out description line */
	while (!feof (f))
	{
		fgets (buf, sizeof (buf), f);
		numroutes++;
	}
	numroutes--;

	routes = (routerec *) malloc (numroutes * sizeof (routerec));

	rewind (f);
	
	fgets (buf, sizeof (buf), f); 
	for (i = 0; i < numroutes; i++)
	{
		if (fgets (buf, sizeof (buf), f) == NULL)
		{
		/* Important, since an interface might have been removed since our counting,
				causing us to parse bogus data */
			fputs ("Error reading /proc/net/route: iface count mismatch\n", stderr);
			fclose (f);
			exit (EXIT_FAILURE);
		}
		if ( strlen (buf) == sizeof(buf)-1 )
		{
		/* skip long lines */
			fputs ("Long (corrupt) line encountered, skipping.\n", stderr);
			while ((fgets (buf, sizeof (buf), f)))
				if (buf [strlen (buf) - 1] == '\n')
					break;
			continue; /* continue with next regular line (or fail if EOF */
		}
		ptr = strtok (buf, "\t ");
		if (!ptr)
			continue;

		if (strlen (ptr) >= sizeof (iface))
			continue; /* would overflow if fed with bogus data in a chroot()ed environment */
		else
			strcpy (iface, ptr);
		ptr = strtok (NULL, "\t ");       /* hack avoiding fscanf */
		routes[i].addr.s_addr=(unsigned long)strtoul(ptr,NULL,16);
		for (i1 = 0; i1 < 6; i1++)
		{
			ptr = strtok (NULL, "\t ");   /* ignore Gateway Flags RefCnt Use Metric */
		}
		if (!ptr) 
		{
			fputs ("Error parsing /proc/net/route\n", stderr);
			continue;
		}
		routes[i].naddr=(unsigned long)strtoul(ptr,NULL,16);   /* Netmask */

		found = 0;
		for (i1 = 0; i1 < numinterfaces; i1++)
		{
			if (strcmp (interfaces[i1].ifname, iface) == 0)
			{
				routes[i].iface = &interfaces[i1];
				found = 1;
			}

		}

		if (!found)
		{
			printf ("Couldn't find interface %s\n", iface);
			exit (EXIT_FAILURE);
		}
	}
	fclose (f);
}
